'use strict';

const axios = require("axios"),
    authUtils = require("http-auth-utils"),
    Q = require("q"),
    crypto = require("crypto"),
    authorization = require("auth-header");

module.exports = {
    /**
     * Starts to read the stream the "success" callback is called on every frame
     * @param success(base64)
     * @param error(error)
     * @param url
     * @param user
     * @param password
     * @param errorCount
     */
    startStream(success, error, url, user, password, errorCount) {
        let probeAbortController = new AbortController();
        let _errorCount = errorCount || 0;
        this._digestCounter = 0;
        this._isDigestAuth = false;
        this._isBasicAuth = false;

        this._successCallback = b64String => {
            _errorCount = 0;
            if (this._readyForFrame) {
                this._readyForFrame = false;
                success(b64String);
            }
        };
        this._errorCallback = (errorUrl, err) => {
            _errorCount++;
            if (_errorCount >= 3 || err.message === 'canceled') {
                this.stopStream();
                error(err);
            } else {
                this.startStream(success, error, url, user, password, _errorCount);
            }
        };

        this._getAuthHeader({
            url: url,
            user: user,
            pwd: password
        }).then(authValue => {
            let config = {
                responseType: "stream",
                signal: probeAbortController.signal
            };
            if (authValue) {
                config.headers = config.headers || {};
                config.headers["Authorization"] = authValue;
            }
            if (this._isBasicAuth) {
                config.auth = {
                    username: user,
                    password: password
                }
            }

            // Validate the contentType to prevent us from downloading a MJPEG stream instead of JPG
            axios.get(url, config).then((resp) => {
                probeAbortController.abort();
                probeAbortController.signal.onabort = () => {
                    resp.data.req.abort();
                    resp.data.req.destroy();
                };
                let contentType = resp.headers["content-type"];
                this.stopStream();
                this._abortHandler = new AbortController();
                if (contentType.startsWith("multipart/x-mixed-replace")) {
                    this._fetchMjpgFrame({
                        url: url,
                        config: config
                    });
                } else if (contentType.includes("image/jpeg")) {
                    this._fetchJpg({
                        url: url,
                        config: config
                    });
                } else {
                    this._errorCallback(url, `Unsupported contentType: ${contentType}`);
                }
            }, err => {
                if (typeof err === "object" && err.message === "Canceled") {
                    return; // Ignore, its due to the probeAbortController
                }
                this._errorCallback(url, err.message)
            });
        }, err => {
            this._errorCallback(url, err);
        });
    },

    _getAuthHeader({url, user, pwd}) {
        let abortController = new AbortController(),
            authValue = null,
            resolveDef = Q.defer(),
            basicAuthValue;

        axios.get(url, {
            responseType: "stream",
            signal: abortController.signal
        }).then(resp => {
            abortController.abort();
            abortController.signal.onabort = () => {
                resp.data.req.abort();
                resp.data.req.destroy();
            };

            if (resp.headers && resp.headers["content-type"] && resp.headers["content-type"].startsWith("multipart/x-mixed-replace")) {
                this.stopStream();
                this._abortHandler = new AbortController();
                this._processMjpgData(resp);

                this._abortHandler.signal.onabort = () => {
                    resp.data.req.abort();
                    resp.data.req.destroy();
                };
            } else {
                resolveDef.resolve(authValue);
            }
        }, err => {
            let { status, headers } = err.response,
                authHeader,
                authMechanism;

            if (headers && headers["www-authenticate"]) {
                try {
                    // this parse method may fail, so wrap it into a try catch
                    authHeader = authUtils.parseWWWAuthenticateHeader(headers["www-authenticate"]);
                } catch {
                    try {
                        // fallback for the first one
                        authHeader = authUtils.parseAuthorizationHeader(headers["www-authenticate"]);
                    } catch {
                        try {
                            let tmpHeader =  authorization.parse(headers["www-authenticate"]);

                            authHeader = {
                                type: tmpHeader.scheme,
                                data: tmpHeader.params
                            };
                        } catch {

                        }
                        // if this also fails, the auth headers are invalid
                    }
                }
                authMechanism = authUtils[authHeader.type.toUpperCase()];
            } else {
                return resolveDef.resolve();
            }

            if (status === 401) {
                if (authMechanism) {
                    if (authHeader.type === "Digest") {
                        this._isDigestAuth = true;
                        this.realm = authHeader.data.realm;
                        this.method = "GET";
                        this.nonce = authHeader.data.nonce;
                        this.cnonce = this._createDigestCnonce();
                        this.opaque = authHeader.data.opaque;

                        try {
                            this.algorithm = authHeader.data.algorithm.toUpperCase();
                        } catch (e) {
                            this.algorithm = authHeader.type.toUpperCase();
                        }
                        this.qop = authHeader.data.qop;
                        this.nonceCount = this._getDigestCounterValue();
                        this.uri = err.request.path;
                        this.user = user;
                        this.pwd = pwd;
                        this.errorStatusText = err.response.statusText;

                        resolveDef.resolve(this._createDigestAuthString());
                    } else if (authHeader.type === "Basic") {
                        this._isBasicAuth = true;
                        resolveDef.resolve();
                    }
                } else {
                    resolveDef.reject(`Authentication method not supported: ${authHeader.type}`);
                }
            }
            resolveDef.resolve(authValue);
        });

        return resolveDef.promise;
    },

    _createHeaderEntry(name, value, last, noQuotes) {
        let field;

        if (!value) {
            return "";
        }

        if (noQuotes) {
            field = name + "=" + value + (last ? "" : ", ");
        } else {
            field = name + "=\"" + value + "\"" + (last ? "" : ", ");
        }

        return field;
    },

    _getDigestCounterValue() {
        return ("00000000" + this._digestCounter).slice(-8);
    },

    _createDigestCnonce() {
        return crypto.randomBytes(7).toString('hex').substring(0, 13);
    },

    _adoptDigestAuthHeader(headers) {
        this._digestCounter++;
        this.nonceCount = this._getDigestCounterValue();

        headers[Object.keys(headers).find(key => key.toLowerCase() === "authorization")] = this._createDigestAuthString();

        return headers;
    },

    _createDigestAuthString() {
        let ha1, ha2, response;

        ha1 = crypto.createHash('md5').update(`${this.user}:${this.realm}:${this.pwd}`).digest('hex');
        if (this.algorithm && this.algorithm.toLowerCase() === "MD5-sess".toLowerCase()) {
            ha1 = crypto.createHash('md5').update(`${ha1}:${this.nonce}:${this.cnonce}`).digest('hex');
        }
        if (this.qop && this.qop.toLowerCase() === 'auth-int') {
            ha2 = crypto.createHash('md5').update(`${this.method}:${this.uri}:${this.errorStatusText}`).digest('hex');
        } else {
            ha2 = crypto.createHash('md5').update(`${this.method}:${this.uri}`).digest('hex');
        }
        if (this.qop && this.qop.toLowerCase() === 'auth') {
            response = crypto.createHash('md5').update(`${ha1}:${this.nonce}:${this.nonceCount}:${this.cnonce}:${this.qop}:${ha2}`).digest('hex');
        } else {
            response = crypto.createHash('md5').update(`${ha1}:${this.nonce}:${ha2}`).digest('hex');
        }

        let result = "Digest ";
        result += this._createHeaderEntry("username", this.user);
        result += this._createHeaderEntry("realm", this.realm);
        result += this._createHeaderEntry("nonce", this.nonce);
        result += this._createHeaderEntry("uri", this.uri);
        result += this._createHeaderEntry("algorithm", this.algorithm);
        result += this._createHeaderEntry("qop", this.qop, false, true);
        result += this._createHeaderEntry("nc", this.nonceCount, false, true);
        result += this._createHeaderEntry("cnonce", this.cnonce);

        if (this.opaque && this.opaque.length) {
            result += this._createHeaderEntry("response", response);
            result += this._createHeaderEntry("opaque", this.opaque, true);
        } else {
            result += this._createHeaderEntry("response", response, true);
        }

        return result;
    },

    /**
     * Stops the currently active stream
     */
    stopStream() {
        this._abortHandler && this._abortHandler.abort();
        this.imageProcessed();
    },

    /**
     * Call this when the UI is ready processing the image,
     * no further frame will be emitted to the callback until this method is called
     */
    imageProcessed() {
        this._readyForFrame = true;
    },

    /**
     * Reads the MJPEG stream and extracts every frame until VideoProxy.stopStream() is called
     * @param url
     * @param headers
     * @private
     */
    _fetchMjpgFrame({url, config}) {
        let headers,
            axiosConfig = {
                responseType: "stream",
                signal: this._abortHandler.signal
            };

        if (this._isDigestAuth) {
            headers = this._adoptDigestAuthHeader(config.headers);
            axiosConfig.headers = headers;

        }  else if (this._isBasicAuth) {
            axiosConfig.auth = config.auth;
        }

        axios.get(url, axiosConfig).then(resp => {
            this._processMjpgData(resp);
            this._abortHandler.signal.onabort = () => {
                resp.data.req.abort();
                resp.data.req.destroy();
            };
        }, this._errorCallback.bind(this, url));
    },

    /**
     * Fetches JPG images until VideoProxy.stopStream() is called
     * @param url
     * @param headers
     * @private
     */
    _fetchJpg({url, config}) {
        let headers,
            axiosConfig = {
                responseType: "arraybuffer",
                signal: this._abortHandler.signal
            };

        if (this._isDigestAuth) {
            headers = this._adoptDigestAuthHeader(config.headers);
            axiosConfig.headers = headers;

        } else if (this._isBasicAuth) {
            axiosConfig.auth = config.auth
        }

        axios.get(url, axiosConfig).then(resp => {
            this._fetchJpg({url, config});
            this._successCallback(Buffer.from(resp.data).toString('base64'));
        }, this._errorCallback.bind(this, url));
    },

    _processMjpgData(resp) {
        var chunks = [],
            soi = Buffer.from([0xff, 0xd8]), // Start Of Input
            eoi = Buffer.from([0xff, 0xd9]); // End Of Input

        resp.data.on("data", chunk => {
            if (chunk.indexOf(soi) !== -1) { // Check if the data chunk contains the JPEG start magic
                chunks = [];
                chunks.push(chunk.slice(chunk.indexOf(soi), chunk.length));
            } else {
                if (chunk.indexOf(eoi) === -1) {  // The data chunk doesn't contain the JPEG end magic, add it to the received chunks
                    chunks.push(chunk);
                } else { // The data chunk contains the JPEG end magic, extract the data until  the JPEG end magic and send the buffer to the successCallback
                    chunks.push(chunk.slice(0, chunk.indexOf(eoi) + 1 + 1)); // +1 is for the second bit of the eoi buffer, the other +1 is for get the length instead of the index
                    this._successCallback(Buffer.concat(chunks).toString("base64"));
                }
            }
        });
    }
};
